#!/usr/bin/perl
#
# VMH - Vim Mail Handler
# New versions available at http://www.fleiner.com/vmh/
# Send emails regarding this code to vmh@fleiner.com
#
# Distributed in the hope that it is useful
#
# Copyright (C) 1998  Claudio Fleiner <claudio@fleiner.com>
#
# WARNING: this is work in progress, it is not yet complete, has several
# bugs and, if you use it, can make you lose emails, send emails to
# the wrong people and upset your sysadmin. US IT AT YOUR OWN RISK
#
# See the file README for further details

use FileHandle;

#initialization (get the variables and read the initialization files)
# Those variables can all be overwritten in the $HOME/.vmhrc file

# The users home directory (used to check for .vmhrc)
$home=$ENV{'HOME'};

# The directory where the Mail is stored (usually $HOME/Mail)
$mail="$home/Mail";

# The home directory of VMH
$vmh=$ENV{'VMH'};

# The editor to use (must be vim version 5.3, others have not been tested)
$editor="vim";

# The directory were the vmh executable are stored (mostly perl scripts)
$exec=$vmh;

# the sendmail executable
$sendmail="/usr/lib/sendmail";

# regexps of addresses not to inlucde in CC when replying
$nocc = "cfl\@zurich.ibm.com|cfl\@almaden.ibm.com|claudio\@fleiner.com|fleiner\@icsi.berkeley.edu";

# change to the Mail directory and read the startup script
chdir "$mail" || die "vmh: cannot chdir to $mail: $!\n";

# sorting criteria (always with respect to email addresses)
# This is used for inbox, other folders are sorted by date/replies
# Emails that do not belong into a mailing list are added at the 
# beginning if they have been addresses directly to you (To: or CC:)
# or at the very end if they were not (those are usually either SPAM
# or mailing lists not listed here)
# Note that the mailinglist strings are regexp!
# The lines tested are 'To:', 'Cc:', 'From:' and 'Sender:'
# Instead of only using this for email lists you can also list
# company names like .ibm.com. Emails from those domains will then
# be grouped together. A very handy feature if you get many emails
# from one or two companies.
@mailinglists = ( 
      'swiss-list@lists\.Stanford\.EDU',
      'linux-security@redhat\.com',
      'vim.*@vim\.org',
      'BEATCALC@aol\.com',
      'owner-linux-kernel@vger\.rutgers\.edu',
      'gtk-list@redhat\.com',
      'gnupg-devel@gnupg.org',
      'www\.tages-anzeiger\.ch@ICSI\.Berkeley\.EDU',
      'GERMNEWS@LISTSERV\.GMD\.DE',
      'owner-hwaytalk-digest@mail\.rapidsite\.net',
      'SC22JSG-request@dkuug\.dk',
      'owner-h323implementors@imtc\.org'
);

# list your email addresses. Emails that have not been sent
# from a mailing list will be shown at the beginning if they have
# been addressed to one of the addresses below
@youremail = ( 'claudio@fleiner.com',
               'fleiner@acm.org' );

sub old_encode_base64 ($;$)
{
    my $res = "";
    my $eol = $_[1];
    $eol = "\n" unless defined $eol;
    pos($_[0]) = 0;                          # ensure start at the beginning
    while ($_[0] =~ /(.{1,45})/gs) {
	$res .= substr(pack('u', $1), 1);
	chop($res);
    }
    $res =~ tr|` -_|AA-Za-z0-9+/|;               # `# help emacs
    # fix padding at the end
    my $padding = (3 - length($_[0]) % 3) % 3;
    $res =~ s/.{$padding}$/'=' x $padding/e if $padding;
    # break encoded string into lines of no more than 76 characters each
    if (length $eol) {
	$res =~ s/(.{1,76})/$1$eol/g;
    }
    $res;
}


sub old_decode_base64 ($)
{
    local($^W) = 0; # unpack("u",...) gives bogus warning in 5.00[123]

    my $str = shift;
    my $res = "";

    $str =~ tr|A-Za-z0-9+=/||cd;            # remove non-base64 chars
    if (length($str) % 4) {
	require Carp;
	Carp::croak("Base64 decoder requires string length to be a multiple of 4")
    }
    $str =~ s/=+$//;                        # remove padding
    $str =~ tr|A-Za-z0-9+/| -_|;            # convert to uuencoded format
    while ($str =~ /(.{1,60})/gs) {
	my $len = chr(32 + length($1)*3/4); # compute length byte
	$res .= unpack("u", $len . $1 );    # uudecode
    }
    $res;
}

# Encode the current window with MIME 64
sub encode_win64() {
   $b=$curwin->Buffer();
   $c=$b->Count();
   undef %header;
   $current="";
   $res="";
   for($i=1;$i<=$c;$i++) {
      $res.=$b->Get($i)."\n";
   }
   $res=&old_encode_base64($res);
   VIM::DoCommand("%d");
   &vimlet("@@",$res);
   VIM::DoCommand("normal P");
   return ;
}

sub dirs() {
   local($topdir,$subdir)=@_;
   local($tdir,@sd);
   $tdir="";
   if(!($topdir eq "")) {
      $tdir=$topdir."/";
   }
   if(!($subdir eq "")) {
      $tdir.=$subdir;
      push @dirs,$tdir;
      chdir $subdir || die "vmh: dirs: cannot chdir to $subdir: $!\n";
   } else {
      $subdir=".";
   }
   opendir(DIR,".") || die "vmh: cannot open directory $subdir: $!\n";
   @sd = grep { /^[^.]/ && -d "$_" } readdir(DIR);
   closedir DIR;

   foreach $i (@sd) { &dirs($tdir,$i);	 }
   !($subdir eq ".") && chdir "..";
}

sub scandirs() {
   local($i);
   undef @dirs;
   VIM::Msg("rescanning directory structure");
   &dirs(); 
   open(D, ">.vmh_dirs") || die "vmh: cannot open .vmh_dirs for writing: $!\n";
   foreach $i (sort(@dirs)) {
      print D "$i\n";
   }
   close D;
}

sub readdirs() {
   open(DR, ".vmh_dirs") || scandirs();
   @dirs=<DR>;
   close DR;
   @dirs=grep { chop($_); } @dirs;
}

sub create_vmh() {
   local($dirname)=@_;
   local(@msg,%fn);
   local($subject,$i,$gotheader,$from);
   VIM::Msg("Reading $dirname");

   open FN,"$dirname/.vmh_cache" || die "vmh: cannot open $fn/.vmh_cache for reading: $!";
   while(<FN>) {
      chop;
      /^.. *(\d+)/;
      $fn{$1}=$_;
   }
   close FN;
   $new=" ";
   $new="*" if($dirname eq "$mail/inbox");


   opendir(DIR,$dirname) || die "vmh: cannot open directory $dirname: $!\n";
   @msg = sort { $a <=> $b} grep { /^[0-9]*$/ && -f "$dirname/$_" } readdir(DIR);
   closedir DIR;
   if($dirname eq "$mail/inbox") {
      $new="*" if($dirname eq "$mail/inbox");
      open VMH,"| sort -f -t '|' +0.1 -0.2 +0.29 -1.0 +0.2 -0.6 > $dirname/.vmh_cache" || die "vmh: cannot open .vmh_cache for writing: $!\n";
   } else {
      $new=" ";
      open VMH,">$dirname/.vmh_cache" || die "vmh: cannot open .vmh_cache for writing: $!\n";
   }
#open VMH,">$dirname/.vmh_cache" || die "vmh: cannot open .vmh_cache for writing: $!\n";
   format_name VMH "CACHE";
   foreach $i (@msg) {
      $cur=$new;

      $move="";
      if(!$fn{$i} eq '') {
	 $fn{$i} =~ /^(.).*>([^>]*)$/;
	 $cur=$1;
	 $move=$2;
      }
      open MSG,"$dirname/$i" || die "vmh: cannot open message $i: $!\n";
      $gotheader=0;
      $subject="";
      $from="";
      $email="";
      $sort="@";
      $emailcont=0;
      while(<MSG>) {
	 chop;
	 if(/^$/ && $gotheader==0) { 
	    $gotheader=1; 
	    if(length($subject)<60) {
	       $subject.=" | ";
	    }
	 }
	 elsif(/^Subject: *(.*)$/ && $gotheader==0) { $subject=$1; }
	 elsif(/^From: *.*\((.*)\)$/ && $gotheader==0) { $from=$1; }
	 elsif(/^From: *"?([^"]*[^ "])"? *<.*>$/ && $gotheader==0) { $from=$1; }
	 elsif(/^From: *(.*)$/ && $gotheader==0) { $from=$1; }
         if((($emailcont==1 && /^[ \t]/) || /^(From|To|Cc|Reply|Sender|Errors-To): (.*)/i) && $gotheader==0) {
            $emailcont=1;
            $email.=" ".$_;
         } else {
            $emailcont=0;
         }
	 if($gotheader) { 
	    if(length($subject)<60) {
	       $subject.=$_;
	    } else {
	       last;
	    }
	 }
      }

      if($dirname eq "$mail/inbox") {
         foreach $j (@youremail) {
            if($email =~ /$j/i) {
               $sort="!";
               last;
            }
         }

         if($sort eq "@") {
            $sort=chr(126);
            for($j=0;$j<=$#mailinglists;$j++) {
               if($email =~ /$mailinglists[$j]/i) {
                  $sort=chr($j+65);
                  last;
               }
            }
         }
      } else {
         $sort=" ";
      }
      $answer=" ";
      if($subject =~ /^(Re|Aw):[ \t]*(.*)/i) {
         $answer="`";
         $subject=$2;
      }

      write VMH;
#print VMH "$email\n";
      close MSG;
   }
   close VMH;

}

format CACHE = 
@@@>>> @<<<<<<<<<<<<<<<<<<<<@@<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<  >@<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$cur,$sort,$i,$from,        $answer,$subject,                                   $move
.

sub getCacheBuffer() {
   local(@buf,$i);
   @buf=VIM::Buffers();
   foreach $i (@buf) {
      return VIM::Buffers($i) if(VIM::Buffers($i)->Name() eq ".vmh_cache");
   }
   die "cannot find .vmh_cache";
}


sub newdir() {
   local($name)=@_;
   local($found,$count);

   $found=VIM::Buffers();
   &create_vmh("$mail/$name");

   if($found!=2) {
      VIM::DoCommand("e $name/.vmh_cache");
      VIM::DoCommand("e!");
   } else {
      VIM::DoCommand("split $name/.vmh_cache");
   }
}

sub dirdefmapping() {
   local($what)=@_;
   VIM::DoCommand("nmap <CR> <esc>g") if $what==2;
   VIM::DoCommand("nmap <CR> <esc>m") if $what==1;
}

sub getdirname() {
   local($dir);
   $dir=$curwin->Buffer->Name();
   $dir=~s+/[^/]*$++;
   VIM::DoCommand("let dirname=\"$dir\"");
}

sub newnumber() {
   local($dir,@sd)=@_;
   opendir(DIR,$dir) || die "vmh: cannot open directory $dir: $!\n";
   @sd = sort { $b <=> $a} grep { /^[0-9]+$/ && -f "$dir/$_" } readdir(DIR);
   closedir DIR;
   if($#sd==-1) {
      return 1;
   } 
   return $sd[0]+1;
}

sub newdraftnumber() {
   local($dir,@sd)=@_;
   opendir(DIR,$dir) || die "vmh: cannot open directory $dir: $!\n";
   @sd = sort { $b <=> $a} grep { /^draft.([0-9]+)$/ && -f "$dir/$_"; $_=$1 } readdir(DIR);
   closedir DIR;
   if($#sd==-1) {
      return "draft.1";
   } 
   return "draft.".($sd[0]+1);

}

sub packdir() {
   local($fn,@fl,$i,$j,$msg,$mmsg,$newdir);
   VIM::Msg("packing dir");
   $fn=$curwin->Buffer->Name();
   $dir=$fn;
   $dir=~s+/[^/]*$++;
   VIM::Msg("working in directory $dir");

   $vmhcache=0;
   if(open FN,"$dir/.vmh_cache") { 
      $vmhcache=1;
      while(<FN>) {
         chop;
         /^.. *(\d+)/;
         $fn{$1}=$_;
      }
      close FN;
   }

   opendir(DIR,$dir) || die "vmh: cannot open directory $dir: $!\n";
   @msg = sort { $a <=> $b} grep { /^[0-9]*$/ && -f "$dir/$_" } readdir(DIR);
   closedir DIR;

   $mmsg=1;
   foreach $i (@msg) {
      VIM::Msg("renaming $i --> $mmsg");
      if($i!=$mmsg) { 
	 link("$dir/".$i,"$dir/".$mmsg);
	 unlink("$dir/".$i);
         if($fn{$i} ne "") {
            $fn{$i}=~s/$i/$mmsg/;
         }
      }
      $mmsg++;
   }
   if($vmhcache) {
      if(open FN,">$dir/.vmh_cache") { 
         foreach $i (values %fn) {
            print FN "$i\n";
         }
         close FN;
      }
   }
   &create_vmh("$mail/$dir");
}

sub commit() {
   local($fn,@fl,$i,$j,$msg,$mmsg,$newdir);
   VIM::Msg("Committing");
   $fn=$curwin->Buffer->Name();
   $dir=$fn;
   $dir=~s+/[^/]*$++;
   open FN,"$fn" || die "vmh: cannot open $fn for reading: $!";
   @fl=<FN>;
   close FN;

# find the lines that end in '>D$'
   foreach $i (@fl) {
      chop $i;
      if($i =~ / >D *$/) {
	 $i =~ /^.. *(\d+)/;
	 unlink "$dir/,$1";
	 link "$dir/$1","$dir/,$1";
	 unlink "$dir/$1";
      }
   }

#moving
   foreach $i (@fl) {
      if($i =~ / > *([A-Z-a-z0-9.][A-Z0-9a-z\/._-]+)$/) {
	 $newdir=$1;
	 $i =~ /^.. *(\d+)/;
	 $msg=$1;
         &createdir($newdir) unless -d $newdir;
	 if(! -d $newdir) {
	    VIM::Msg("unknwon directory $newdir");
	 } else {
	    $mmsg=&newnumber($newdir);
	    &movefile("$dir/$msg","$newdir/$mmsg");
	 }
      }
   }
   &create_vmh("$mail/$dir");
}

sub createdir() {
   local($new)=@_;
   local($dir,@d,$i);
   @d=split(/\//,$new);
   $dir=".";
   foreach $i (@d) {
      $dir.="/".$i;
      mkdir($dir,0700) unless -d $dir;
   }
}

sub movefile() {
   local($old,$new)=@_;
   if(link "$old","$new") {
      unlink "$old";
   } else {
      open O,"$old" || die "vmh: cannot open $old for reading: $!";
      open N,">$new" || die "vmh: cannot open $new for writing: $!";
      while(<O>) {
	 print N $_;
      }
      close O;
      close N;
      unlink "$old";
   }
}

sub getheaders() {
   local($b,$c,$t,$i,$current);
   $b=$curwin->Buffer();
   $c=$b->Count();
   undef %header;
   $current="";
   for($i=1;$i<=$c;$i++) {
      $l=$b->Get($i);
      if($l =~ /^$/) { return 0; }
      if($l =~ /^[ \t]+(.*)/) { 
	 $header{$current}.="\n\t".$1;
      } elsif($l =~ /^([a-z])([a-z-]+): +(.*)/i) {
	 ($current=$1) =~ tr [a-z] [A-Z];
	 ($t=$2) =~ tr [A-Z] [a-z];
	 $current.=$t;
	 $header{$current}.=$3;
      }
   }
   return -1;
}

sub sendemail() {
   local($b,$c,$i,$from,$to,$cc,$bcc,$subject);
   $i=&getheaders();
   $b=$curwin->Buffer();
   if($1!=0) {
      VIM::Msg("vmh: Illegal header line $i") if $i>0;
      VIM::Msg("vmh: no body found") if $i<0;
      return;
   }
   foreach $i ('From', 'To', 'Subject') {
      if($header{$i} eq "") {
	 VIM::Msg("vmh: no '$i' line defined");
	 return;
      }
   }

#remove empty headers
   for($i=1;$i<=$b->Count();) {
      $l=$b->Get($i);
      last if($l =~ /^$/);
      if($l =~ /^[A-Za-z0-9-]+:[ \t]*$/) {
	 $n=$b->Get($i+1);
	 if($n =~ /^$/ || $n =~ /^[^ \t]/) {
	    $b->Delete($i);
	 } else {
	    $i++;
	 }
      } else {
	 $i++;
      }
   }

   VIM::Msg("starting sendmail");
   if(!open(SENDMAIL,"|$sendmail -i -t")) {
      VIM::Msg("vmh: cannot execute sendmail");
   } else {
      for($i=1;$i<=$b->Count();$i++) {
	 $l=$b->Get($i);
# VIM::Msg("sending $l");
	 print SENDMAIL "$l\n";
      }
      close SENDMAIL;
      VIM::Msg("sendmail done");

      $mmsg=&newnumber("sent");
      unlink $b->Name() || VIM::Msg("cannot unlink ".$b->Name());
      VIM::DoCommand(":w sent/$mmsg");
      VIM::DoCommand(":bd");
      VIM::Msg("mail has been sent and stored in \"sent/$mmsg\"");
   }
}

sub newemail() {
   VIM::DoCommand("let vmh_newemail='".&newdraftnumber("drafts")."'");
}

sub vimlet() {
   local($n,$v)=@_;
   $v=~s/\\/\\\\/g;
   $v=~s/"/\\"/g;
   VIM::DoCommand("let $n=\"$v\"");
}
sub replyemail() {
   local($sub);
   &getaddress();
   $sub=$header{'Subject'};
   if($sub !~ /^[ \t]*[Rr][Ee]:/) {
      $sub="Re: ".$sub;
      &vimlet("vmh_subject",$sub);
   }
}

sub getaddress() {
   local($b,$c,$i,$from,$to,$cc,$bcc,$subject,@cc);
   $i=&getheaders();

   if($header{'Reply-to'} eq "") {
      &vimlet("vmh_to",$header{'From'});
      $v="";
   } else {
      &vimlet("vmh_to",$header{'Reply-to'});
      $v=$header{'From'};
   }
   $v.=", ".$header{'To'} if($header{'To'} ne "");
   $v.=", ".$header{'Cc'} if($header{'Cc'} ne "");
   print "*************************$v\n";
   (@cc)=split(/,[, \t\n]*/,$v);
   $v="";
   $b=0;
   foreach $c (@cc) {
      if($c !~ $nocc) {
         if($b+length($c)>70) {
            $b=0;
            $v.=", \n\t";
         } else {
            $v.=", " if($v ne "");
         }
         $b+=length($c);
         $v.=$c;
      }
   }
   &vimlet("vvv",$v);
   &vimlet("vmh_cc",$v);
   &vimlet("vmh_subject",$header{'Subject'});
   &vimlet("vmh_date",$header{'Date'});
   &vimlet("vmh_messageid",$header{'Message-id'});
   &vimlet("vmh_from",$header{'From'});
   VIM::DoCommand("let vmh_newemail='".&newdraftnumber("drafts")."'");
}


&readdirs();
VIM::DoCommand("split");
VIM::DoCommand("view .vmh_dirs");
VIM::DoCommand("hide");
VIM::DoCommand("split address");
#VIM::DoCommand("hide");

&newdir("inbox");
1;
